------------------------------------------------------------
-- Copyright (c) 2016 tacigar. All rights reserved.
------------------------------------------------------------
-- Copyright (c) 2020 IFRFSX.
------------------------------------------------------------
-- Copyleft (Я) 2021-2022 mazes
-- https://gitlab.com/mazes_80/maidroid
------------------------------------------------------------

local S = maidroid.translator

local timers = maidroid.timers

-- Core interface functions
local on_start, on_pause, on_resume, on_stop, on_step, is_tool, is_alttool

-- Core extra functions
local plant, mow, to_action
local craft_seeds, select_seed, task, task_base
local is_seed, is_plantable, is_mowable, is_scythe

local wander_core =  maidroid.cores.wander
local to_wander = wander_core.to_wander
local core_path = maidroid.cores.path

local search = maidroid.helpers.search_surrounding

local mature_plants = {}
local seeds = {}

local farming_redo = farming and farming.mod and farming.mod == "redo"

if farming_redo then
	local mature, crop
	for _, v in pairs(farming.registered_plants) do
		--[IFRFSX] replace k to v.crop,this is plant's real name.
		mature = v.crop .. "_" .. v.steps
		crop = v.crop .. "_1"
		mature_plants[mature] = {chance=1, crop=crop}
		seeds[v.seed] = v.crop .. "_1"
	end
	-- Pepper can be harvested when green or yellow too
	mature_plants["farming:pepper_7"] = {chance=1, crop="farming:pepper_1"}
	mature_plants["farming:pepper_6"] = {chance=3, crop="farming:pepper_1"}
	mature_plants["farming:pepper_5"] = {chance=9, crop="farming:pepper_1"}
	if not maidroid.mods.sickles then
		maidroid.register_tool_rotation("farming:scythe_mithril", vector.new(-75,45,-45))
	end
else
	local plant_list = { "cotton", "wheat" }
	local crop
	for _, plantname in ipairs(plant_list) do
		crop = "farming:" .. plantname .. "_1"
		mature_plants["farming:" .. plantname .. "_8"] = {chance=1,crop=crop}
		seeds["farming:seed_" .. plantname] = crop
	end
end

local ethereal_plants={}
if maidroid.mods.ethereal then
	-- [plant_name_without_steps]= { seed = "seed_name", step = int_step }
	ethereal_plants["ethereal:strawberry"] = {seed = "ethereal:strawberry", step = 8}
	ethereal_plants["ethereal:onion"] = {seed = "ethereal:wild_onion_plant", step = 5}

	for name, def in pairs(ethereal_plants) do
		local mature = name .. "_" .. def.step
		local crop = name .. "_1"
		mature_plants[mature] = {chance=1,crop=crop}
		seeds[def.seed] = crop
	end
end

local cucina_vegana_mod = maidroid.mods.cucina_vegana
if cucina_vegana_mod then
	local germ = "1"
	if farming_redo then
		germ = tonumber(cucina_vegana.plant_settings.germ_launch)
		if germ == 0 then
			germ = "seed"
		end
	end

	for _, val in ipairs(cucina_vegana.plant_settings.bonemeal_list) do
		-- val1: plantname_, val[2]: steps, val[3]: seed
		local crop
		if farming_redo then
			crop = val[1] .. germ
		else
			crop = val[3]
		end
		seeds[val[3]] = crop
		mature_plants[val[1] .. val[2]] = {chance=1,crop=crop}
	end
end

-- is_plantable reports whether maidroid can plant any seed.
is_plantable = function(pos, name)
	if minetest.is_protected(pos, name) then
		return false
	end
	local node = minetest.get_node(pos)
	local lpos = vector.add(pos, {x = 0, y = -1, z = 0})
	local lnode = minetest.get_node(lpos)
	return node.name == "air"
		and minetest.get_item_group(lnode.name, "soil") > 1
end

-- is_mowable reports whether maidroid can mow.
is_mowable = function(pos, name)
	if minetest.is_protected(pos, name) then
		return false
	end

	local node = minetest.get_node(pos)
	local desc = mature_plants[node.name]

	if desc == nil then
		return false
	end
	return math.random(desc.chance) == 1
end

local supports = {}
if farming_redo then
	supports["farming:beans"] = "farming:beanpole"
	supports["farming:grapes"] = "farming:trellis"
end

-- select_seed select the first available seed stack
select_seed = function(self)
	local support

	for _, stack in pairs(self:get_inventory():get_list("main")) do
		if not stack:is_empty() and is_seed(stack:get_name()) then
			support = supports[stack:get_name()]
			if not support or self:get_inventory():contains_item("main", support) then
				self.selected_seed = stack:get_name()
				return
			end
		end
	end

	self.selected_seed = craft_seeds(self) -- May be nil during a full step

	if not self.selected_seed then
		self.need_core_selection = true
		return
	end

	if self.state ~= maidroid.states.WANDER then
		to_wander(self)
	end
end

on_start = function(self)
	self.path = nil
	wander_core.on_start(self)
end

on_resume = function(self)
	self.path = nil
	select_seed(self)
	wander_core.on_resume(self)
end

on_stop = function(self)
	self.path = nil
	wander_core.on_stop(self)
end

on_pause = function(self)
	wander_core.on_pause(self)
end

local position_ok = function(pos, to)
	local dist = vector.distance(pos, to)
	if dist < 1 then return true end -- Always good

	local from = vector.round(vector.copy(pos))
	-- Node directly under must be pumpkin or watermelon
	if to.x == from.x and to.z == from.z and to.y == from.y - 1 then
		return true
	end

	-- Skip when inside block
	local node = minetest.get_node(from)
	node = minetest.registered_nodes[node]
	if  node and node.buildable_to and
		( ( from.y == to.y and dist < 1.41 ) or
		( from.y ~= to.y and dist < 1.73 ) ) then
		return true
	end

	return false
end

task_base = function(self, comparator, action)
	local pos = self:get_pos()
	local destination = search(pos, comparator, self.owner)
	if destination == nil then
		return
	end

	-- Do we act on the target
	if position_ok(pos, destination) then
		self.destination = destination
		self.action = action
		to_action(self)
		return true
	end

	-- Do we need to follow a path
	local path = minetest.find_path(pos, destination, 8, 1, 1, "A*_noprefetch")
	if path ~= nil then
		core_path.to_follow_path(self, path, destination, to_action, action)
		return true
	end
end

task = function(self)
	-- Planting
	if self.selected_seed and is_seed(self.selected_seed) and
		task_base(self, is_plantable, plant) then
		return
	end
	-- Harvesting
	task_base(self, is_mowable, mow)
end

is_seed = function(name)
	return seeds[name] ~= nil
end

local seed_recipes = {}
if farming_redo then
	-- Notice: Keep seed and replacement switched for pineapple
	seed_recipes["farming:garlic"]        = { count = 8, seed = "farming:garlic_clove" }
	seed_recipes["farming:melon_8"]       = { count = 4, seed = "farming:melon_slice" }
	seed_recipes["farming:pepper"]        = { count = 1, seed = "farming:peppercorn" }
	seed_recipes["farming:pepper_yellow"] = { count = 1, seed = "farming:peppercorn" }
	seed_recipes["farming:pepper_red"]    = { count = 1, seed = "farming:peppercorn" }
	seed_recipes["farming:pineapple"]     = { count = 5, seed = "farming:pineapple_ring" }
	seed_recipes["farming:pumpkin_8"]     = { count = 4, seed = "farming:pumpkin_slice" }
	seed_recipes["farming:sunflower"]     = { count = 5, seed = "farming:seed_sunflower" }

	seed_recipes["farming:melon_8"].tool          = "farming:cutting_board"
	seed_recipes["farming:pineapple"].replacement = "farming:pineapple_top"
	seed_recipes["farming:pumpkin_8"].tool        = "farming:cutting_board"
end

-- Maximal stuff used for recipe
for plantname, rules in pairs(seed_recipes) do
	local stack = ItemStack(plantname)
	local max = stack:get_stack_max()
	rules.remove = math.floor(max / rules.count)
end

craft_seeds = function(self)
	local inv = self:get_inventory()
	local stack, name, count, rules

	-- Does inventory contains any valid recipe to make seeds
	for _, inv_stack in pairs(inv:get_list("main")) do
		name = inv_stack:get_name()
		rules = seed_recipes[name]

		if rules and ( not rules.tool or
			inv:contains_item("main", rules.tool) ) then
			break
		end
	end
	if not rules then return end -- No recipe matched

	-- Remove used things from inventory
	stack = inv:remove_item("main", ItemStack(name .. " " .. rules.remove))
	count = stack:get_count()

	if count == 0 then return false end -- Nothing was used

	-- Prepare output and store in inventory
	stack = { ItemStack(rules.seed .. " " .. count * rules.count) }
	if rules.replacement then
		table.insert(stack, ItemStack(rules.replacement .. " " .. count))
	end
	self:add_items_to_main(stack)
	-- NOTICE highly care about this when add new seed_recipes
	return rules.replacement or rules.seed
end

to_action = function(self)
	self.state = maidroid.states.ACT
	self.timers.action = 0
	self.timers.walk = 0
end

local place_plant_support = function(self, plantname)
	local support = supports[plantname]
	if not support then
		return true
	end

	local inv = self:get_inventory()
	local stack = ItemStack(support)
	stack = inv:remove_item("main", stack)
	if stack:get_count() == 0 then
		return false
	end
	return true
end

local freeze_action = function(self, toolname)
	-- Check tool was not removed from inventory
	if not toolname or not self:get_inventory():contains_item("main", toolname) then
		self.need_core_selection = true
		return true
	end
	self:set_tool(toolname)
	self:halt()
	self:set_animation(maidroid.animation.MINE)
	self:set_yaw({self:get_pos(), self.destination})
end

local update_action_timers = function(self, dtime, toolname)
	if self.timers.action < timers.action_max then
		if self.timers.action == 0 then
			if freeze_action(self, toolname) then
				return true
			end
		end
		self.timers.action = self.timers.action + dtime
		return true
	end
	return false
end

plant = function(self, dtime)
	-- Skip until timer is ok
	if update_action_timers(self, dtime, self.selected_seed) then return end

	if	not self.selected_seed or
		not place_plant_support(self, self.selected_seed) then
		self.need_core_selection = true
		return
	end

	-- We can place plant
	local inv = self:get_inventory()
	local plantname = seeds[self.selected_seed] -- Lookup for a crop
	minetest.set_node(self.destination, { name = plantname,
		param2 = minetest.registered_nodes[plantname].place_param2 or 1 } )
	minetest.sound_play("default_place_node", {pos = self.destination, gain = 1.0})
	inv:remove_item("main", ItemStack(self.selected_seed))

	-- Update maidroid state
	if not inv:contains_item("main", self.selected_seed) then
		self.need_core_selection = true
		return -- Last selected seed used
	end
	to_wander(self, 0, timers.change_dir_max )
	self:set_tool(self.selected_tool)
end

mow = function(self, dtime)
	-- TODO play dig sound
	-- Skip until timer is ok
	if update_action_timers(self, dtime, self.selected_tool) then return end

	local destnode = minetest.get_node(self.destination)
	local mature = destnode.name
	local stacks

	if is_scythe(self.selected_tool) then -- Fast tool for farmers
		local name = mature_plants[mature].crop
		local p2 = minetest.registered_nodes[name].place_param2 or 1
		local filters = mature
		stacks = {}
		if farming_redo and mature:sub(1,-2) == "farming:pepper_" then
			filters = { "farming:pepper_5", "farming:pepper_6", "farming:pepper_7" }
		end -- Treat pepper as a special case
		local nodes = minetest.find_nodes_in_area(
			vector.add(self.destination, {x=-1,y=-1,z=-1}),
			vector.add(self.destination, {x=1, y=1, z=1}),
			filters
		) -- Find connected nodes matching this mature plant

		local count = 0
		local ok = false
		local drops
		for _, pos in ipairs(nodes) do
			if ok and count == 4 then -- Scythes treats 5 plants at most
				break
			end
			if	ok or -- Target node already harvested
				count < 4 or -- Slot still available for target
				pos == self.destination then -- Always for target
				if filters ~= mature then -- This pepper is hot
					drops = minetest.get_node_drops(minetest.get_node(pos).name)
				else
					drops = minetest.get_node_drops(mature)
				end
				table.insert_all(stacks, drops) -- Save the drops
				minetest.set_node(pos, { name = name, param2 = p2 } )
				count = count + 1
			end
			if pos == self.destination then
				count = count - 1
				ok = true
			end
		end
	else -- Normal mode
		stacks = minetest.get_node_drops(mature)
		minetest.remove_node(self.destination)
	end

	self:add_items_to_main(stacks)
	to_wander(self, 0, timers.change_dir_max )
end

on_step = function(self, dtime, moveresult)
	self:toggle_entity_jump(dtime, moveresult)

	-- When owner offline mode disabled and if owner didn't login, the maidroid does nothing.
	if maidroid.settings.farming_offline == false
		and not minetest.get_player_by_name(self.owner) then
		return
	end

	-- Pickup surrounding items
	self:pickup_item()

	if self.state == maidroid.states.WANDER then
		wander_core.on_step(self, dtime, moveresult, task, maidroid.criteria.is_fence, true)
	elseif self.state == maidroid.states.PATH then
		maidroid.cores.path.on_step(self, dtime, moveresult)
	elseif self.state == maidroid.states.ACT then
		self.action(self, dtime)
	end
end

is_scythe = function(name)
	return name == "farming:scythe_mithril"
		or minetest:get_item_group(name, "scythe") > 0
end

is_tool = function(stack)
	local name = stack:get_name()
	return minetest.get_item_group(name, "hoe") > 0
		or is_scythe(name)
end

is_alttool = function(stack)
	return is_seed(stack:get_name())
end

maidroid.register_core("farming", {
	description	= S("a farmer"),
	on_start	= on_start,
	on_stop		= on_stop,
	on_resume	= on_resume,
	on_pause	= on_pause,
	on_step		= on_step,
	is_tool		= is_tool,
	select_tool	= select_seed,
	is_alttool  = is_alttool,
	walk_max = 2.5 * timers.walk_max,
})
maidroid.new_state("ACT")

-- vim: ai:noet:ts=4:sw=4:fdm=indent:syntax=lua
